#include <config.h>

#include <stdio.h>
#ifndef __USE_BSD
# define __USE_BSD	/* Needed for mkdtemp () */
#endif

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <strings.h>

#ifdef ENABLE_NLS
# include <libintl.h>
#endif

#include "bmpx/util.h"
#include "bmpx/uri.h"
#include "bmpx/vfs.h"
#include "bmpx/plugin.h"
#include "bmpx/plugin_interfaces.h"

#include "libcontainer_m3u.h"

#ifdef HAVE_FTS_H
#  include <fts.h>
#endif

static VFSHandle *handle = NULL;

gboolean
container_m3u_can_process (BmpPluginContainerM3U *self, const gchar *uri)
{
    if (str_has_suffix_nocase (uri, ".m3u")) 
	return TRUE;
    return FALSE;
}


gboolean
container_m3u_handle_set (BmpPluginContainerM3U *self, VFSHandle *_handle)
{
    if (_handle == NULL) 
	return FALSE;
    
    if (!_handle->valid)
	return FALSE;

    handle = _handle;

    return TRUE;
}

VFSHandle*
container_m3u_handle_get (BmpPluginContainerM3U *self)
{
    return handle;
}

gboolean
container_m3u_handle_remove (BmpPluginContainerM3U *self)
{
    handle = NULL;

    return TRUE;
}

gboolean 
container_m3u_handle_read (BmpPluginContainerM3U *self)
{
    GList  *uri_list = NULL;
    gchar **uris;
    gint n = 0;

    g_return_val_if_fail (handle != NULL, FALSE);
    g_return_val_if_fail (VFS_HANDLE_VALID(handle), FALSE);

    if ((!handle->buffer) || (!strlen(handle->buffer)))    
      {
	handle->uri_list = NULL;
	return TRUE;
      }

    uris =
      g_strsplit (handle->buffer, "\n", 0);

    while (uris[n])
	{
	    gchar *uri = uris[n];

	    /* Strip leading spaces/tabs */
	    while ((uri[0] == ' ') || (uri[0] == '\t'))
		uri++;
	    /* Skip comments, extended information or null lenght uris */
	    if ((uri[0] == '#') || (uri[0] == '\0'))
		{
		    n++;
		    continue;
		}
	    uri_list = g_list_append (uri_list, g_strdup (uri));
	    n++;
	}

    g_strfreev (uris);

    handle->uri_list = uri_list;
    return TRUE;
}

static void
container_m3u_interface_init (gpointer         g_iface,
                              gpointer         iface_data)
{
        BmpPluginTransportInterfaceClass *klass = (BmpPluginTransportInterfaceClass *)g_iface;

	klass->can_process    = container_m3u_can_process;
	klass->handle_set     = container_m3u_handle_set;
	klass->handle_get     = container_m3u_handle_get;
	klass->handle_remove  = container_m3u_handle_remove;
	klass->handle_read    = container_m3u_handle_read;
}

/* Metadata IFace */
static gchar* _metadata[] = {
    "M3U",
    "Milosz Derezynski",
    "(C) GPL 2005",
    "http://bmpx.berlios.de",
    "internalerror@gmail.com",
};

static const gchar**
get_metadata (BmpPlugMetadataInterface *self)
{
    return _metadata;
}

static void
metadata_init (gpointer         g_iface,
               gpointer         iface_data)
{
        BmpPlugMetadataInterfaceClass *klass = (BmpPlugMetadataInterfaceClass *)g_iface;

        klass->get_metadata = get_metadata;
}

GType plugin_container_m3u_type = 0;

void 
plugin_container_m3u_register_type (GTypeModule *module)
{
	    static const GTypeInfo object_info = {
                        sizeof (BmpPluginContainerM3UClass),
                        NULL,   /* base_init */
                        NULL,   /* base_finalize */
                        NULL,   /* class_init */
                        NULL,   /* class_finalize */
                        NULL,   /* class_data */
                        sizeof (BmpPluginContainerM3U),
                        0,      /* n_preallocs */
                        NULL    /* instance_init */
                };

                plugin_container_m3u_type = g_type_module_register_type (module,
					     G_TYPE_OBJECT,
					     "BmpPluginContainerM3UType",
					     &object_info, 0);

                static const GInterfaceInfo interface_container_m3u_info = {
                        (GInterfaceInitFunc) container_m3u_interface_init,    /* interface_init */
                        NULL,                                       /* interface_finalize */
                        NULL                                        /* interface_data */
                };

                static const GInterfaceInfo interface_metadata_info = {
                        (GInterfaceInitFunc) metadata_init,    /* interface_init */
                        NULL,                                       /* interface_finalize */
                        NULL                                        /* interface_data */
                };

                g_type_module_add_interface (module,
					     plugin_container_m3u_type,
                                             BMP_TYPE_PLUG_METADATA,
                                             &interface_metadata_info);


                g_type_module_add_interface (module,
					     plugin_container_m3u_type,
                                             BMP_TYPE_PLUGIN_CONTAINER_IFACE,
                                             &interface_container_m3u_info);
}
