/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*- */

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>

#include <glib.h>
#include <glib-object.h>

#include "bmpx/plugin.h"
#include "bmpx/playlist.h"
#include "bmpx/metadata.h"
#include "bmpx/list.h"
#include "bmpx/plugin_interfaces.h"

#include "libflow_shuffle.h"

/* Shuffle Flow Control */
static BmpPlaybackHistoryItem*
flow_shuffle_item_next (BmpPluginFlowShuffle *self, const BmpListType *list, gint _idx)
{
    BmpPlaylistItem *playlist;
    BmpFileItem *file_item;
    BmpPlaybackHistoryItem *history_item;
    gint length;
    gint current;
    gint track;

    playlist = 
	BMP_PLAYLIST_ITEM(g_list_nth_data(list->items, _idx)); 

    length =
	g_list_length (playlist->list->items)-1;

    current =
	playlist->list->current;

    history_item = 
	g_new0 (BmpPlaybackHistoryItem,1);

    history_item->list_time = 
	    playlist->list->time;

    history_item->list_name = 
	    g_strdup (playlist->listname);

    g_random_set_seed((guint32)time(NULL));

    track = current; 

    while (track == current)    
      {
	track = 
	    g_random_int_range (0, length); 
      }
    
    file_item = BMP_FILE_ITEM(g_list_nth_data(playlist->list->items, track));
    history_item->guid = g_malloc0(sizeof(RowGUID));
    memcpy (history_item->guid, file_item->guid, sizeof(RowGUID));

    return history_item;
}

static BmpPlaybackHistoryItem*
flow_shuffle_item_prev  (BmpPluginFlowShuffle *self, const BmpListType *list, gint _idx)
{
    BmpPlaylistItem *playlist;
    BmpFileItem *file_item;
    BmpPlaybackHistoryItem *history_item;
    gint length;
    gint current;
    gint track;

    playlist = 
	BMP_PLAYLIST_ITEM(g_list_nth_data(list->items, _idx)); 

    length =
	g_list_length (playlist->list->items)-1;

    current =
	playlist->list->current;

    history_item = 
	g_new0 (BmpPlaybackHistoryItem,1);

    history_item->list_time = 
	    playlist->list->time;

    history_item->list_name = 
	    g_strdup (playlist->listname);

    g_random_set_seed((guint32)time(NULL));

    track = current; 

    while (track == current)    
      {
	track = 
	    g_random_int_range (0, length); 
      }

    file_item = BMP_FILE_ITEM(g_list_nth_data(playlist->list->items, track));
    history_item->guid = g_malloc0(sizeof(RowGUID));
    memcpy (history_item->guid, file_item->guid, sizeof(RowGUID));

    return history_item;
}

/* IFaces */
static gchar* _metadata[] = {
    "Shuffle",
    "Milosz Derezynski",
    "(C) GPL 2005",
    "http://bmpx.berlios.de",
    "internalerror@gmail.com",
};

static const gchar**
get_metadata (BmpPlugMetadataInterface *self)
{
    return _metadata;
}

static void
metadata_init (gpointer         g_iface,
	       gpointer         iface_data)
{
	BmpPlugMetadataInterfaceClass *klass = (BmpPlugMetadataInterfaceClass *)g_iface;

	klass->get_metadata = get_metadata;
}


void
flow_shuffle_interface_init (gpointer         g_iface)
{
        BmpPluginFlowInterfaceClass *klass = (BmpPluginFlowInterfaceClass *)g_iface;

        klass->item_next = flow_shuffle_item_next;
        klass->item_prev = flow_shuffle_item_prev;
}


GType plugin_flow_shuffle_type = 0;

void 
plugin_flow_shuffle_register_type (GTypeModule *module)
{
	    static const GTypeInfo object_info = {
                        sizeof (BmpPluginFlowShuffleClass),
                        NULL,   /* base_init */
                        NULL,   /* base_finalize */
                        NULL,   /* class_init */
                        NULL,   /* class_finalize */
                        NULL,   /* class_data */
                        sizeof (BmpPluginFlowShuffle),
                        0,      /* n_preallocs */
                        NULL    /* instance_init */
                };

                static const GInterfaceInfo interface_flow_shuffle_info = {
                        (GInterfaceInitFunc) flow_shuffle_interface_init,    /* interface_init */
                        NULL,                                       /* interface_finalize */
                        NULL                                        /* interface_data */
                };

                static const GInterfaceInfo interface_metadata_info = {
                        (GInterfaceInitFunc) metadata_init,    /* interface_init */
                        NULL,                                       /* interface_finalize */
                        NULL                                        /* interface_data */
                };

                plugin_flow_shuffle_type = g_type_module_register_type (module,
					   G_TYPE_OBJECT,
					  "BmpPluginFlowShuffleType",
					   &object_info, 0);

                g_type_module_add_interface (module,
					     plugin_flow_shuffle_type,
                                             BMP_TYPE_PLUG_METADATA,
                                             &interface_metadata_info);

                g_type_module_add_interface (module,
					     plugin_flow_shuffle_type,
                                             BMP_TYPE_PLUGIN_FLOW_IFACE,
                                             &interface_flow_shuffle_info);
}
