/*  BMP - Cross-platform multimedia player
 *  Copyright (C) 2003-2004  BMP development team.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public Licensse as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <config.h>
#include <revision.h>
#include <build.h>

#include <glib.h>
#include <glib/gi18n.h>
#include <gtk/gtk.h>
#include <glade/glade.h>
#include <string.h>
#include <limits.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>

#include "main.h"
#include "glade.h"

#include "bmp_ui.h"
#include "ui_main.h"
#include "ui_playlist.h"
#include "ui_prefswin.h"
#include "chroma_list.h"
#include "plugin_priv.h"

#include "bmp_tooltips.h"

#include "bmpx/util.h"

#define THUMBNAIL_WIDTH  90
#define THUMBNAIL_HEIGHT 40

enum SkinViewCols {
    SKIN_VIEW_COL_PREVIEW,
    SKIN_VIEW_COL_NAME,
    SKIN_VIEW_COL_PATH,
    SKIN_VIEW_N_COLS
};

enum CategoryViewCols {
    CATEGORY_VIEW_COL_ICON,
    CATEGORY_VIEW_COL_NAME,
    CATEGORY_VIEW_COL_ID,
    CATEGORY_VIEW_COL_ICON_ACTIVE,
    CATEGORY_VIEW_COL_ICON_INACTIVE,

    CATEGORY_VIEW_N_COLS
};

enum PluginViewCols {
    PLUGIN_VIEW_COL_ACTIVE,
    PLUGIN_VIEW_COL_DESC,
    PLUGIN_VIEW_COL_FILENAME,
    PLUGIN_VIEW_COL_ID,
    PLUGIN_VIEW_N_COLS
};

#define SKIN_NODE(x)  ((SkinNode *)(x))
struct _SkinNode {
    gchar *name;
    gchar *path;
    GTime *time;
};
typedef struct _SkinNode SkinNode;

typedef struct {
    const gchar *icon_path;
    const gchar *icon_path_inactive;
    const gchar *name;
    gint id;
} Category;

typedef struct {
    const gchar *name;
    const gchar *tag;
}
TitleFieldTag;

static Category categories[] = {
    {DATA_DIR "/images/appearance.png", DATA_DIR "/images/appearance_ia.png", N_("Appearance"), 0},
    {DATA_DIR "/images/eq.png",         DATA_DIR "/images/eq_ia.png",         N_("Equalizer"), 1},
    {DATA_DIR "/images/library.png",    DATA_DIR "/images/library_ia.png",    N_("Library"), 5},
    {DATA_DIR "/images/mouse.png",      DATA_DIR "/images/mouse_ia.png",      N_("Mouse"), 2},
    {DATA_DIR "/images/playlist.png",   DATA_DIR "/images/playlist_ia.png",   N_("Playlist"), 3},
    {DATA_DIR "/images/playback.png",   DATA_DIR "/images/playback_ia.png",   N_("Playback"), 4},
    {DATA_DIR "/images/plugins.png",    DATA_DIR "/images/plugins_ia.png",    N_("Plugins"), 6}
};
static gint n_categories = G_N_ELEMENTS(categories);

static BmpTooltipEntry window_preferences_tooltips[] = {
    { "titlestring_entry",
      N_("Allows you to customize how text is displayed in the playlist"),
      GTK_STOCK_INFO },

    { "skin_view",
      N_("Here you can choose a skin to determine how BMP looks like"),
      GTK_STOCK_INFO },

    { "font_per_skin",
      N_("Enabling this will make BMP remember the font for each skin individually, giving you more ease when theming BMP"),
      GTK_STOCK_INFO },

    { "font_button",
      N_("Click on this button to select a font for BMP to use"),
      GTK_STOCK_INFO }
};
static gint n_window_preferences_tooltips = G_N_ELEMENTS (window_preferences_tooltips);


static gchar* eq_sliders[] = {
    "eq_30",
    "eq_60",
    "eq_125",
    "eq_250",
    "eq_500",
    "eq_1000",
    "eq_2000",
    "eq_4000",
    "eq_8000",
    "eq_16000"
};
static gint n_eq_sliders = G_N_ELEMENTS(eq_sliders);

static void
update_skinlist(BmpWindowPreferences *self);

static GdkPixbuf *
skin_get_thumbnail(const gchar * path);

/* GObject stuff */
G_DEFINE_TYPE (BmpWindowPreferences, bmp_window_preferences, G_TYPE_OBJECT)

void
window_preferences_create(BmpWindowPreferences *self);

BmpWindowPreferences*
bmp_window_preferences_new (void)
{
    BmpWindowPreferences *self;

    self =
        g_object_new (BMP_TYPE_WINDOW_PREFERENCES, NULL);

    return self;
}

static GObject *
bmp_window_preferences_constructor (GType type,
	                     guint n_construct_properties,
         	             GObjectConstructParam * construct_properties)
{
    GObject *obj;

    {
        /*
         * Invoke parent constructor.
         */
        BmpWindowPreferencesClass *klass;
        GObjectClass *parent_class;
        klass = BMP_WINDOW_PREFERENCES_CLASS (g_type_class_peek (BMP_TYPE_WINDOW_PREFERENCES));
        parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
        obj =
            parent_class->constructor (type,
                                       n_construct_properties,
                                       construct_properties);
    }

    /*
     * do stuff.
     */
    return obj;
}

static void
bmp_window_preferences_class_init (BmpWindowPreferencesClass * klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

#if 0
    gobject_class->set_property = bmp_play_set_property;
    gobject_class->get_property = bmp_play_get_property;
    gobject_class->dispose = bmp_play_dispose;
    gobject_class->finalize = bmp_play_finalize;
#endif
    gobject_class->constructor = bmp_window_preferences_constructor;

}

struct _BmpWindowPreferencesPrivate {

    gboolean dispose_has_run;

    GladeXML  *xml;
    gchar *skins_path;
    GList *skinlist;
    GtkWidget *skin_view;
    GtkWidget *skin_view_notebook;
};


static void
bmp_window_preferences_init (BmpWindowPreferences *self)
{
    self->private = g_new (BmpWindowPreferencesPrivate, 1);

    self->private->dispose_has_run = FALSE;
    self->private->skins_path = NULL;
    self->private->skinlist = NULL;

    window_preferences_create(self);
}

/* Our stuff */
static void
change_category(GtkNotebook * notebook,
                GtkTreeSelection * selection)
{
    GtkTreeModel *model;
    GtkTreeIter iter;
    gint index;

    if (!gtk_tree_selection_get_selected(selection, &model, &iter))
        return;

    gtk_tree_model_get(model, &iter, CATEGORY_VIEW_COL_ID, &index, -1);
    gtk_notebook_set_current_page(notebook, index);
}

static void
on_titlestring_entry_changed (GtkComboBoxEntry *widget, gpointer data)
{
    bmp_configuration_key_set_string (bmp_configuration,
                                      "bmp",
                                      "titlestring-format",
                                      gtk_combo_box_get_active_text (GTK_COMBO_BOX(widget)));
}

static void
on_font_button_font_set(GtkFontButton * button,
                                 gpointer data)
{
    const gchar *fontname;

    fontname = gtk_font_button_get_font_name (button);
    bmp_configuration_key_set_string (bmp_configuration, "bmp", "font", fontname);

    /* Now the font-per-skin handling */
    if (bmp_configuration_key_fetch_boolean (bmp_configuration, "bmp", "font-per-skin"))
    {
	    gchar *path, *basename, *basename_archive;

	    bmp_configuration_domain_new (bmp_configuration, "skin-fonts");

	    path = bmp_configuration_key_fetch (bmp_configuration, "bmp", "skin");
	    basename = g_path_get_basename (path);
	    basename_archive = archive_basename (basename);

	    bmp_configuration_key_set_string (bmp_configuration, "skin-fonts", basename_archive, fontname);
    }
    else
    {
	    bmp_configuration_key_set_string (bmp_configuration, "bmp", "font-save", fontname);
    }
}

void
on_font_button_realize(GtkFontButton * button,
                                gpointer data)
{
    gchar *aux;

    aux = bmp_configuration_key_fetch (bmp_configuration, "bmp", "font");
    gtk_font_button_set_font_name(button, aux);
    g_free (aux);
}

static void
on_playlist_show_pl_numbers_realize(GtkToggleButton * button,
                                    gpointer data)
{
#if 0
    gtk_toggle_button_set_active(button, cfg.show_numbers_in_pl);
#endif
}

static void
on_playlist_show_pl_numbers_toggled(GtkToggleButton * button,
                                    gpointer data)
{
#if 0
    cfg.show_numbers_in_pl = gtk_toggle_button_get_active(button);
#endif
}

static void
on_custom_cursors_realize(GtkToggleButton *button,
                          gpointer data)
{
    LOOKUP_OR_CREATE_DEFAULT_BOOLEAN("bmp","use-custom-cursors", TRUE);
    gtk_toggle_button_set_active(button, bmp_configuration_key_fetch_boolean (bmp_configuration, "bmp","use-custom-cursors"));
}

static void
on_custom_cursors_toggled(GtkToggleButton *button,
                          gpointer data)
{
    BmpWindowMain *bmp_main;
    BmpWindowPlaylist *playlist;

    bmp_configuration_key_set_boolean (bmp_configuration, "bmp", "use-custom-cursors", gtk_toggle_button_get_active(button));

    bmp_main = RM_GET_ITEM("windows","window-main", NOLOCK);
    playlist = RM_GET_ITEM("windows","window-playlist", NOLOCK);

    bmp_window_main_configure (bmp_main);
    bmp_window_playlist_configure (playlist);
}

static void
on_display_tooltips_realize(GtkToggleButton *button,
                          gpointer data)
{
    gboolean value;

    LOOKUP_OR_CREATE_DEFAULT_BOOLEAN("bmp","display-tooltips", TRUE);

    value = bmp_configuration_key_fetch_boolean (bmp_configuration, "bmp", "display-tooltips");
    gtk_toggle_button_set_active (button, value);
}

static void
on_display_tooltips_toggled(GtkToggleButton *button,
                          gpointer data)
{
    BmpTooltips *tooltips;
    gboolean value;

    value = gtk_toggle_button_get_active (button);
    bmp_configuration_key_set_boolean (bmp_configuration, "bmp", "display-tooltips", value);

    tooltips = RM_GET_ITEM ("tooltips","tooltips-preferences", LOCK);
    if (! value)
	bmp_tooltips_disable (tooltips);
    else
	bmp_tooltips_enable  (tooltips);
    RM_ITEM_UNLOCK("tooltips","tooltips-preferences");

    tooltips = RM_GET_ITEM ("tooltips","tooltips-main", LOCK);
    if (! value)
	bmp_tooltips_disable (tooltips);
    else
	bmp_tooltips_enable  (tooltips);
    RM_ITEM_UNLOCK("tooltips","tooltips-main");
}

static void
on_display_numbers_realize(GtkToggleButton *button,
                          gpointer data)
{
    gboolean value;

    LOOKUP_OR_CREATE_DEFAULT_BOOLEAN("bmp","display-tracklist-numbers", TRUE);
    value = bmp_configuration_key_fetch_boolean (bmp_configuration, "bmp", "display-tracklist-numbers");

    gtk_toggle_button_set_active (button, value);
}

static void
on_display_numbers_toggled(GtkToggleButton *button,
                          gpointer data)
{
    BmpWindowPlaylist *playlist;
    ChromaList *list;
    gboolean value;

    value = gtk_toggle_button_get_active (button);
    bmp_configuration_key_set_boolean (bmp_configuration, "bmp", "display-tracklist-numbers", value);

    playlist = RM_GET_ITEM("windows","window-playlist", NOLOCK);
    list = CHROMA_LIST (chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget);

    chroma_list_set_use_numbers (list, value);
    gtk_widget_queue_draw (playlist->window);
}

static void
on_font_per_skin_realize(GtkToggleButton *button,
                          gpointer data)
{
    gboolean value;

    LOOKUP_OR_CREATE_DEFAULT_BOOLEAN("bmp","font-per-skin",FALSE);

    value = bmp_configuration_key_fetch_boolean (bmp_configuration, "bmp", "font-per-skin");
    gtk_toggle_button_set_active (button, value);
}

static void
on_font_per_skin_toggled(GtkToggleButton *button,
                         BmpWindowPreferences *self)
{
    gchar *path, *basename, *basename_archive, *skin_font;
    gboolean value;

    value = gtk_toggle_button_get_active (button);
    bmp_configuration_key_set_boolean (bmp_configuration, "bmp", "font-per-skin", value);

    if (value == TRUE)
    {
	gchar *aux;

	aux = bmp_configuration_key_fetch (bmp_configuration, "bmp", "font-save");
	if (! aux)
	{
	    aux = bmp_configuration_key_fetch (bmp_configuration, "bmp", "font");
	    bmp_configuration_key_set_string (bmp_configuration, "bmp", "font-save", aux);
	    g_free (aux);
	}
	else
	    g_free (aux);

	bmp_configuration_domain_new (bmp_configuration, "skin-fonts");

	path = bmp_configuration_key_fetch (bmp_configuration, "bmp", "skin");
	basename = g_path_get_basename (path);
	basename_archive = archive_basename (basename);

	skin_font = bmp_configuration_key_fetch (bmp_configuration, "skin-fonts", basename_archive);

	if (skin_font)
	{
	    bmp_window_preferences_set_font (self, skin_font);
	    bmp_configuration_key_set_string (bmp_configuration, "bmp", "font", skin_font);
	    g_free (skin_font);
	}
	else
	{
	    skin_font = bmp_configuration_key_fetch (bmp_configuration, "bmp", "font-save");

	    if (skin_font)
	    {
		bmp_window_preferences_set_font (self, skin_font);
		bmp_configuration_key_set_string (bmp_configuration, "bmp", "font", skin_font);
		g_free (skin_font);
	    }
	}
    }
    else
    {
	skin_font = bmp_configuration_key_fetch (bmp_configuration, "bmp", "font-save");
	if (skin_font)
	{
	    bmp_window_preferences_set_font (self, skin_font);
	    bmp_configuration_key_set_string (bmp_configuration, "bmp", "font", skin_font);
	    g_free (skin_font);
	}
    }

}


static void
skin_view_update(BmpWindowPreferences *self)
{
    GtkTreeView *treeview = GTK_TREE_VIEW(self->private->skin_view);
    GtkTreeSelection *selection = NULL;
    GtkListStore *store;
    GtkTreeIter iter, iter_current_skin;
    GtkTreePath *path;
    GdkPixbuf *thumbnail;
    gchar *name, *current_skin, *basename = NULL, *current_skin_basename = NULL;
    GList *entry;

    current_skin = bmp_configuration_key_fetch (bmp_configuration, "bmp", "skin");
    if (current_skin)
    {
	basename = g_path_get_basename (current_skin);
	current_skin_basename = archive_basename (basename);
    }

    gtk_notebook_set_current_page (GTK_NOTEBOOK(self->private->skin_view_notebook), 1);

    store = GTK_LIST_STORE(gtk_tree_view_get_model(treeview));
    gtk_list_store_clear(store);

    update_skinlist(self);

    for (entry = self->private->skinlist; entry; entry = g_list_next(entry)) {
        thumbnail = skin_get_thumbnail(SKIN_NODE(entry->data)->path);

        if (!thumbnail)
            continue;

        name = SKIN_NODE(entry->data)->name;

        gtk_list_store_append(store, &iter);
        gtk_list_store_set(store, &iter,
                           SKIN_VIEW_COL_PREVIEW, thumbnail,
                           SKIN_VIEW_COL_NAME, name,
			   SKIN_VIEW_COL_PATH, SKIN_NODE(entry->data)->path, -1);
        g_object_unref(thumbnail);

        if ( current_skin_basename && (!strcmp(current_skin_basename,
                    name))) {
            iter_current_skin = iter;
        }

        while (gtk_events_pending())
            gtk_main_iteration();
    }

    selection = gtk_tree_view_get_selection(treeview);
    gtk_tree_selection_select_iter(selection, &iter_current_skin);

    path = gtk_tree_model_get_path(GTK_TREE_MODEL(store), &iter_current_skin);
    gtk_tree_view_scroll_to_cell(treeview, path, NULL, TRUE, 0.5, 0.5);
    gtk_tree_path_free(path);

    gtk_notebook_set_current_page (GTK_NOTEBOOK(self->private->skin_view_notebook), 0);
}

/* Skins */
static gchar *
get_thumbnail_filename(const gchar * path)
{
    gchar *basename, *pngname, *thumbname;

    g_return_val_if_fail(path != NULL, NULL);

    basename = g_path_get_basename(path);
    pngname = g_strconcat(basename, ".png", NULL);

    thumbname = g_build_filename(BMP_PATH_SKIN_THUMB_DIR,
                                 pngname, NULL);

    g_free(basename);
    g_free(pngname);

    return thumbname;
}


static GdkPixbuf *
skin_get_preview(const gchar * path)
{
    GdkPixbuf *preview = NULL;
    gchar *dec_path, *preview_path;
    gboolean is_archive = FALSE;

    if (file_is_archive(path))
      {
        if (!(dec_path = archive_decompress(path)))
            return NULL;

        is_archive = TRUE;
      }
    else
      {
        dec_path = g_strdup(path);
      }

    preview_path = find_file_recursively(dec_path, "main.bmp");

    if (preview_path)
      {
        preview = gdk_pixbuf_new_from_file(preview_path, NULL);
        g_free(preview_path);
      }
    else
      {
	g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "%s: Unable to find main.bmp in path '%s'", G_STRLOC, dec_path);
      }

    if (is_archive)
        del_directory(dec_path);

    g_free(dec_path);

    return preview;
}


static GdkPixbuf *
skin_get_thumbnail(const gchar * path)
{
    GdkPixbuf *scaled = NULL;
    GdkPixbuf *preview;
    gchar *thumbname;

    g_return_val_if_fail(path != NULL, NULL);

    if (g_str_has_suffix(path, BMP_SKIN_THUMB_DIR_BASENAME))
        return NULL;

    thumbname = get_thumbnail_filename(path);

    if (g_file_test(thumbname, G_FILE_TEST_EXISTS)) {
        scaled = gdk_pixbuf_new_from_file(thumbname, NULL);
        g_free(thumbname);
        return scaled;
    }

    if (!(preview = skin_get_preview(path))) {
        g_free(thumbname);
        return NULL;
    }

    scaled = gdk_pixbuf_scale_simple(preview,
                                     THUMBNAIL_WIDTH, THUMBNAIL_HEIGHT,
                                     GDK_INTERP_BILINEAR);
    g_object_unref(preview);

    gdk_pixbuf_save(scaled, thumbname, "png", NULL, NULL);
    g_free(thumbname);

    return scaled;
}

static void
skinlist_add(BmpWindowPreferences *self, const gchar * filename)
{
    SkinNode *node;
    gchar *basename;

    g_return_if_fail(filename != NULL);

    node = g_new0(SkinNode, 1);
    node->path = g_strdup(filename);

    basename = g_path_get_basename(filename);

    if (file_is_archive(filename))
      {
        node->name = archive_basename(basename);
        g_free(basename);
      }
    else
      {
        node->name = basename;
      }


    self->private->skinlist = g_list_prepend(self->private->skinlist, node);
}

static gboolean
scan_skindir_func(const gchar * path, const gchar * basename, BmpWindowPreferences *self)
{

    if (g_file_test(path, G_FILE_TEST_IS_REGULAR)) {
        if (file_is_archive(path)) {
            skinlist_add(self, path);
        }
    }
    else if (g_file_test(path, G_FILE_TEST_IS_DIR)) {
        skinlist_add(self, path);
    }

    return FALSE;
}

static void
scan_skindir(BmpWindowPreferences *self, const gchar * path)
{
    GError *error = NULL;

    g_return_if_fail(path != NULL);

    if (path[0] == '.')
        return;

    if (!dir_foreach(path, (DirForeachFunc)scan_skindir_func, self, &error)) {
        g_warning("Failed to open directory (%s): %s", path, error->message);
        g_error_free(error);
        return;
    }
}

#if 0
static gint
skinlist_compare_func(gconstpointer a, gconstpointer b)
{
    g_return_val_if_fail(a != NULL && SKIN_NODE(a)->name != NULL, 1);
    g_return_val_if_fail(b != NULL && SKIN_NODE(b)->name != NULL, 1);
    return g_ascii_strcasecmp(SKIN_NODE(a)->name, SKIN_NODE(b)->name);
}
#endif

static void
skin_free_func(gpointer data)
{
    g_return_if_fail(data != NULL);
    g_free(SKIN_NODE(data)->name);
    g_free(SKIN_NODE(data)->path);
    g_free(data);
}

static void
skinlist_clear(BmpWindowPreferences *self)
{
    if (!self->private->skinlist)
        return;

    g_list_foreach(self->private->skinlist, (GFunc) skin_free_func, NULL);
    g_list_free(self->private->skinlist);
    self->private->skinlist = NULL;
}

void
update_skinlist(BmpWindowPreferences *self)
{
    skinlist_clear(self);

    scan_skindir(self, DATA_DIR G_DIR_SEPARATOR_S BMP_SKIN_DIR_BASENAME);
    scan_skindir(self, BMP_PATH_USER_SKIN_DIR);

    g_assert(self->private->skinlist != NULL);
}

static void
skin_view_on_cursor_changed(GtkTreeView * treeview,
                            BmpWindowPreferences *self)
{
    GtkTreeModel *model;
    GtkTreeSelection *selection;
    GtkTreeIter iter;

    gchar *path, *aux;

    selection = gtk_tree_view_get_selection(treeview);
    if (!gtk_tree_selection_get_selected(selection, &model, &iter))
        return;

    gtk_tree_model_get(model, &iter, SKIN_VIEW_COL_PATH, &path, -1);

    aux = bmp_configuration_key_fetch (bmp_configuration, "bmp", "skin");
    if(!strcmp(aux, path))
    {
	g_free (aux);
	return;
    }

    /* This will invoke the signal handler in bmp_ui.c for the configuration's skin path value */
    bmp_configuration_key_set_string (bmp_configuration, "bmp", "skin", path);

    /* Now the font-per-skin handling */
    if (bmp_configuration_key_fetch_boolean (bmp_configuration, "bmp", "font-per-skin"))
    {
	    gchar *basename, *basename_archive, *skin_font;

	    bmp_configuration_domain_new (bmp_configuration, "skin-fonts");
	    basename = g_path_get_basename (path);
	    basename_archive = archive_basename (basename);

	    skin_font = bmp_configuration_key_fetch(bmp_configuration, "skin-fonts", basename_archive);

	    if (skin_font)
	    {
		bmp_window_preferences_set_font (self, skin_font);
		bmp_configuration_key_set_string (bmp_configuration, "bmp", "font", skin_font);
		g_free (skin_font);
	    }
	    else
	    {
		skin_font = bmp_configuration_key_fetch (bmp_configuration, "bmp", "font-save");
		if (skin_font)
		{
		    bmp_window_preferences_set_font (self, skin_font);
		    bmp_configuration_key_set_string (bmp_configuration, "bmp", "font", skin_font);
		    g_free (skin_font);
		}
	    }
	    g_free (basename);
	    g_free (basename_archive);
    }

}

static gint
skin_list_sort_func (GtkTreeModel *model,
                     GtkTreeIter *iter_a,
                     GtkTreeIter *iter_b,
                     gpointer user_data)
{
    gchar *string_a, *string_b;
    gint cmp_val;

    gtk_tree_model_get(model, iter_a, SKIN_VIEW_COL_NAME, &string_a, -1);
    gtk_tree_model_get(model, iter_b, SKIN_VIEW_COL_NAME, &string_b, -1);

    cmp_val = g_ascii_strcasecmp(string_a, string_b);

    return cmp_val;
}

static void
skin_added_func (const gchar *_path, BmpWindowPreferences *self)
{
	GtkListStore *store = GTK_LIST_STORE(gtk_tree_view_get_model(GTK_TREE_VIEW(self->private->skin_view)));
	GdkPixbuf *thumbnail;
	GtkTreeIter iter;
	gchar *name, *path;

	path = g_build_filename (self->private->skins_path, _path, NULL);
	thumbnail = skin_get_thumbnail(path);
       	if (!thumbnail)
		return;
	name = archive_basename(_path);
       	gtk_list_store_append(store, &iter);
	gtk_list_store_set(store, &iter,
       		SKIN_VIEW_COL_PREVIEW, thumbnail,
               	SKIN_VIEW_COL_NAME, name,
		SKIN_VIEW_COL_PATH, path, -1);
	g_object_unref(thumbnail);
#if 0
	g_free (name);
	g_free (path);
#endif
}

static void
skin_removed_func (const gchar *_path, BmpWindowPreferences *self)
{
	GtkTreeModel *store = gtk_tree_view_get_model(GTK_TREE_VIEW(self->private->skin_view));
	GtkTreeIter iter;
	gchar *name, *path;
	gboolean loop = TRUE;

	if (!gtk_tree_model_get_iter_first (store, &iter)) return;
	path = g_build_filename (self->private->skins_path, _path, NULL);

	while (loop)
	{
		gtk_tree_model_get (store, &iter, SKIN_VIEW_COL_PATH, &name, -1);
		if (!strcmp(path,name))
		{
			gtk_list_store_remove(GTK_LIST_STORE(store), &iter);
#if 0
			g_free (path);
			g_free (name);
#endif
			loop = FALSE;
		}
		else
		{
			if (!gtk_tree_model_iter_next (store, &iter)) loop = FALSE;
		}
	}
#if 0
       g_free(path);
#endif
}

static void
skin_view_setup(BmpWindowPreferences *self)
{
    GtkTreeView *treeview = GTK_TREE_VIEW(self->private->skin_view);
    GtkListStore *store;
    GtkTreeViewColumn *column;
    GtkCellRenderer *renderer;
    GtkTreeSelection *selection;

    gtk_widget_show_all(GTK_WIDGET(treeview));

    gtk_tree_view_set_rules_hint(GTK_TREE_VIEW(treeview), TRUE);
    gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(treeview), FALSE);

    store = gtk_list_store_new(SKIN_VIEW_N_COLS, GDK_TYPE_PIXBUF,
                               G_TYPE_STRING, G_TYPE_STRING);
    gtk_tree_view_set_model(treeview, GTK_TREE_MODEL(store));

    column = gtk_tree_view_column_new();
    gtk_tree_view_column_set_sizing(column, GTK_TREE_VIEW_COLUMN_AUTOSIZE);
    gtk_tree_view_column_set_spacing(column, 16);
    gtk_tree_view_append_column(GTK_TREE_VIEW(treeview),
                                GTK_TREE_VIEW_COLUMN(column));

    renderer = gtk_cell_renderer_pixbuf_new();
    gtk_tree_view_column_pack_start(column, renderer, FALSE);
    gtk_tree_view_column_set_attributes(column, renderer, "pixbuf",
                                        SKIN_VIEW_COL_PREVIEW, NULL);
    g_object_set (G_OBJECT(renderer),"xpad",6,NULL);
    g_object_set (G_OBJECT(renderer),"ypad",2,NULL);

    renderer = gtk_cell_renderer_text_new();
    gtk_tree_view_column_pack_start(column, renderer, TRUE);
    gtk_tree_view_column_set_attributes(column, renderer, "text",
                                        SKIN_VIEW_COL_NAME, NULL);

    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeview));
    gtk_tree_selection_set_mode(selection, GTK_SELECTION_BROWSE);

    gtk_tree_sortable_set_sort_func (GTK_TREE_SORTABLE(store),
                                     SKIN_VIEW_COL_NAME,
                                     skin_list_sort_func,
                                     NULL,
                                     NULL);

    gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE(store), SKIN_VIEW_COL_NAME, GTK_SORT_ASCENDING);

    g_signal_connect(treeview, "cursor-changed",
                     G_CALLBACK(skin_view_on_cursor_changed), self);
}

static void
category_view_cell_data_func_pb (GtkTreeViewColumn *tree_column,
                                             GtkCellRenderer *cell,
                                             GtkTreeModel *tree_model,
                                             GtkTreeIter *iter,
                                             GtkTreeView *treeview)
{
    GtkTreeSelection *selection;
    GtkTreePath *path1, *path2;
    GtkTreeIter iter_cmp;
    GdkPixbuf *pb;

    selection = gtk_tree_view_get_selection (treeview);
    if (gtk_tree_selection_get_selected (selection, NULL, &iter_cmp))
    {
	path1 = gtk_tree_model_get_path (tree_model, iter);
        path2 = gtk_tree_model_get_path (tree_model, &iter_cmp);

	if (! gtk_tree_path_compare (path1, path2))
        {
	    gtk_tree_model_get (tree_model, iter, CATEGORY_VIEW_COL_ICON_ACTIVE, &pb, -1);
	    g_object_set (G_OBJECT(cell), "pixbuf", pb, NULL);
        }
	else
        {
	    gtk_tree_model_get (tree_model, iter, CATEGORY_VIEW_COL_ICON_INACTIVE, &pb, -1);
	    g_object_set (G_OBJECT(cell), "pixbuf", pb, NULL);
	}
    }
    else
    {
	    gtk_tree_model_get (tree_model, iter, CATEGORY_VIEW_COL_ICON_INACTIVE, &pb, -1);
	    g_object_set (G_OBJECT(cell), "pixbuf", pb, NULL);
    }

}

static void
category_view_cell_data_func_text (GtkTreeViewColumn *tree_column,
                                   GtkCellRenderer *cell,
                                   GtkTreeModel *tree_model,
                                   GtkTreeIter *iter,
                                   GtkTreeView *treeview)
{
    GtkTreeSelection *selection;
    GtkTreePath *path1, *path2;
    GtkTreeIter iter_cmp;
    gchar *text;

    selection = gtk_tree_view_get_selection (treeview);

    if (gtk_tree_selection_get_selected (selection, NULL, &iter_cmp))
    {
	path1 = gtk_tree_model_get_path (tree_model, iter);
        path2 = gtk_tree_model_get_path (tree_model, &iter_cmp);

	if (gtk_tree_path_compare (path1, path2))
        {
	   g_object_set (G_OBJECT(cell), "foreground", "#909090", NULL);
	}
    }

    gtk_tree_model_get (tree_model, iter, CATEGORY_VIEW_COL_NAME, &text, -1);
    g_object_set (G_OBJECT(cell), "text", text, NULL);
}

static gchar* _plugin_column_names[] = {
	N_("Type"),
	N_("Library"),
	N_("Name"),
	N_("Author"),
	N_("Copyright"),
	N_("Web URL"),
	N_("Email"),
	N_("Type Name"),
	N_("Implements")
};

static void
on_plugins_view_realize (GtkTreeView * treeview,
			 gpointer user_data)
{
    GtkTreeViewColumn   *column;
    GtkCellRendererText *renderer;
    gint n;

    gtk_tree_view_set_headers_visible (treeview, TRUE);

    for (n=0;n<G_N_ELEMENTS(_plugin_column_names);n++)
      {
	column = gtk_tree_view_column_new();
	gtk_tree_view_column_set_title(column, _plugin_column_names[n]);

	renderer = gtk_cell_renderer_text_new();
	gtk_tree_view_column_pack_start(column, renderer, FALSE);
        gtk_tree_view_column_set_attributes(column, renderer, "text", n, NULL);
	gtk_tree_view_append_column(treeview, column);

	if (n == (G_N_ELEMENTS(_plugin_column_names)-1))
	    gtk_tree_view_column_set_expand (column, TRUE);

      }

    gtk_tree_view_set_model (treeview, GTK_TREE_MODEL(plugins_about));
}

static void
on_category_view_realize(GtkTreeView * treeview,
                         GtkNotebook * notebook)
{
    GtkListStore *store;
    GtkCellRenderer *renderer;
    GtkTreeViewColumn *column;
    GtkTreeSelection *selection;
    GtkTreeIter iter;
    GdkPixbuf *img, *img_inactive;
    gint i;

    column = gtk_tree_view_column_new();
#if 0
    gtk_tree_view_column_set_title(column, Q_("Category"));
#endif
    gtk_tree_view_append_column(treeview, column);
    gtk_tree_view_column_set_spacing(column, 2);

    renderer = gtk_cell_renderer_pixbuf_new();
    gtk_tree_view_column_pack_start(column, renderer, FALSE);
    g_object_set (G_OBJECT(renderer),"xpad",2,NULL);
    g_object_set (G_OBJECT(renderer),"ypad",2,NULL);
    gtk_tree_view_column_set_cell_data_func (column, renderer,
                                             (GtkTreeCellDataFunc) category_view_cell_data_func_pb,
                                             treeview, NULL);

    renderer = gtk_cell_renderer_text_new();
    gtk_tree_view_column_pack_start(column, renderer, FALSE);
    g_object_set (G_OBJECT(renderer),"xpad",2,NULL);
    g_object_set (G_OBJECT(renderer),"ypad",2,NULL);
    gtk_tree_view_column_set_cell_data_func (column, renderer,
                                             (GtkTreeCellDataFunc) category_view_cell_data_func_text,
                                             treeview, NULL);

    store = gtk_list_store_new(CATEGORY_VIEW_N_COLS,
                               GDK_TYPE_PIXBUF, G_TYPE_STRING, G_TYPE_INT, GDK_TYPE_PIXBUF, GDK_TYPE_PIXBUF);
    gtk_tree_view_set_model(treeview, GTK_TREE_MODEL(store));

    for (i = 0; i < n_categories; i++) {
        img = gdk_pixbuf_new_from_file(categories[i].icon_path, NULL);
        img_inactive = gdk_pixbuf_new_from_file(categories[i].icon_path_inactive, NULL);
	g_object_ref (img);
	g_object_ref (img_inactive);
        gtk_list_store_append(store, &iter);
        gtk_list_store_set(store, &iter,
                           CATEGORY_VIEW_COL_ICON,
			   img_inactive,
                           CATEGORY_VIEW_COL_NAME,
                           gettext(categories[i].name),
			   CATEGORY_VIEW_COL_ICON_ACTIVE,
			   img,
			   CATEGORY_VIEW_COL_ICON_INACTIVE,
			   img_inactive,
			   CATEGORY_VIEW_COL_ID,
                           categories[i].id,
			   -1);
        g_object_unref(img);
    }

    gtk_tree_view_set_headers_visible (treeview, FALSE);

    selection = gtk_tree_view_get_selection(treeview);

    g_signal_connect_swapped(selection, "changed",
                             G_CALLBACK(change_category), notebook);
}

static void
on_library_add_path_clicked (GtkButton *button, BmpWindowPreferences *self)
{
    GtkWidget *dialog, *dialog_wait;
    GList *items = NULL;
    GError *error = NULL;
    gchar *folder;
    gint rv;

    g_return_if_fail (GTK_IS_BUTTON(button));
    g_return_if_fail (BMP_IS_WINDOW_PREFERENCES(self));

    dialog = gtk_file_chooser_dialog_new (Q_("BMP: Add Path To Library"),
                                          GTK_WINDOW (glade_xml_get_widget_warn (self->private->xml, "preferences_window")),
                                          GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER,
                                          GTK_STOCK_CANCEL,
                                          GTK_RESPONSE_CANCEL,
                                          GTK_STOCK_ADD,
                                          GTK_RESPONSE_OK,
                                          NULL);

    folder = bmp_configuration_key_fetch (bmp_configuration, "bmp", "file-chooser-path");
    gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER(dialog), folder);
    g_free (folder);

    rv = gtk_dialog_run (GTK_DIALOG (dialog));

    if (rv == GTK_RESPONSE_CANCEL)
    {
	gtk_widget_destroy (dialog);
	return;
    }

    folder = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER(dialog));
    gtk_widget_destroy (dialog);

    dialog_wait = glade_xml_get_widget_warn (self->private->xml, "wait_window");
    gtk_window_present (GTK_WINDOW (dialog_wait));

    dir_foreach(folder, (DirForeachFunc)collect_path,
                &items, &error);

    for ( ; items ; items = items->next)
      {
	gchar *uri;

	uri = g_strconcat ("file://", items->data, NULL);
	bmp_metadata_cache_get_metadata_for_uri (bmp_metadata_cache, uri);
	g_free (uri);

	while (gtk_events_pending()) gtk_main_iteration();
      }

    g_list_foreach (g_list_first(items), g_free, NULL);

    bmp_metadata_cache_write (bmp_metadata_cache);

    gtk_widget_hide (dialog_wait);
}

/* FIXME: complete the map */
FUNC_MAP_BEGIN(prefswin_func_map)
    FUNC_MAP_ENTRY(on_custom_cursors_realize)
    FUNC_MAP_ENTRY(on_custom_cursors_toggled)
    FUNC_MAP_ENTRY(on_playlist_show_pl_numbers_realize)
    FUNC_MAP_ENTRY(on_playlist_show_pl_numbers_toggled)
    FUNC_MAP_ENTRY(on_font_per_skin_realize)
    FUNC_MAP_ENTRY(on_font_per_skin_toggled)
    FUNC_MAP_ENTRY(on_display_tooltips_realize)
    FUNC_MAP_ENTRY(on_display_tooltips_toggled)
    FUNC_MAP_ENTRY(on_display_numbers_realize)
    FUNC_MAP_ENTRY(on_display_numbers_toggled)
    FUNC_MAP_ENTRY(on_font_button_font_set)
    FUNC_MAP_ENTRY(on_font_button_realize)
    FUNC_MAP_ENTRY(on_library_add_path_clicked)
    FUNC_MAP_ENTRY(on_plugins_view_realize)
FUNC_MAP_END

static void
on_eq_value_changed (GtkWidget *widget, gpointer user_data)
{
    gdouble value;
    gint value_int;

    g_return_if_fail (GTK_IS_WIDGET(widget));

    value = gtk_range_get_value (GTK_RANGE(widget));
    value_int = (gint)value;

    bmp_system_control_eq_set (bmp_system_control, GPOINTER_TO_INT(user_data), value_int);
}

void
bmp_window_preferences_set_font (BmpWindowPreferences *self, const gchar *font_name)
{
    GtkWidget *font_button;

    font_button = glade_xml_get_widget (self->private->xml, "font_button");
    gtk_font_button_set_font_name(GTK_FONT_BUTTON(font_button), font_name);
}

void
window_preferences_create(BmpWindowPreferences *self)
{
    BmpTooltips *tooltips;
    const gchar *glade_file = DATA_DIR "/glade/dialog_preferences.glade";
    gchar *aux;
    GtkWidget *widget, *widget2;
    gint i;

    /* Initialize tooltips */
    tooltips = bmp_tooltips_new ();
    RM_ITEM_NEW("tooltips", "tooltips-preferences", tooltips, NULL);

    /* load the interface */
    self->private->xml = glade_xml_new_or_die(Q_("Preferences Window"), glade_file, NULL,
                               NULL);

    /* connect the signals in the interface */
    glade_xml_signal_autoconnect_map(self->private->xml, prefswin_func_map, self);
    self->window = glade_xml_get_widget(self->private->xml, "preferences_window");

    /* create category view */
    widget = glade_xml_get_widget(self->private->xml, "category_view");
    widget2 = glade_xml_get_widget(self->private->xml, "category_notebook");
    g_signal_connect_after(G_OBJECT(widget), "realize",
                           G_CALLBACK(on_category_view_realize),
                           widget2);
    self->private->skin_view = glade_xml_get_widget (self->private->xml, "skin_view");
    skin_view_setup (self);
    self->private->skin_view_notebook = glade_xml_get_widget (self->private->xml, "skins_notebook");

    /* Titlestring */
    widget =
	glade_xml_get_widget(self->private->xml, "titlestring_entry");
    LOOKUP_OR_CREATE_DEFAULT("bmp","titlestring-format","<artist> - <album> - <title>");
    aux =
	 bmp_configuration_key_fetch(bmp_configuration,"bmp","titlestring-format");
    gtk_entry_set_text (GTK_ENTRY(GTK_BIN(widget)->child), aux);
    g_free (aux);
    g_signal_connect (G_OBJECT(widget), "changed", G_CALLBACK(on_titlestring_entry_changed), self->window);


    /* Wait images */
    widget = glade_xml_get_widget(self->private->xml, "image_building");
    gtk_image_set_from_file (GTK_IMAGE(widget), DATA_DIR "/images/wait.gif");

    widget = glade_xml_get_widget(self->private->xml, "image_processing");
    gtk_image_set_from_file (GTK_IMAGE(widget), DATA_DIR "/images/wait.gif");

    /* Connect eq sliders */
    for (i = 0; i < n_eq_sliders; i++) {
	widget = glade_xml_get_widget(self->private->xml, eq_sliders[i]);
	g_signal_connect (G_OBJECT(widget), "value-changed", G_CALLBACK(on_eq_value_changed), GINT_TO_POINTER(i));
    }

    for (i = 0; i < n_window_preferences_tooltips; i++) {
	widget = glade_xml_get_widget(self->private->xml, window_preferences_tooltips[i].widget_name);
	bmp_tooltips_set_tip (tooltips,
				widget,
				window_preferences_tooltips[i].stock_id,
				gettext (window_preferences_tooltips[i].tooltip_text),
			        NULL,
				TRUE);
    }

    widget = glade_xml_get_widget(self->private->xml, "label_version");
    gtk_label_set_markup (GTK_LABEL(widget), "<span size=\"small\" color=\"#606060\">BMPx " RV_LAST_CHANGED_DATE "-R" RV_REVISION " [" BUILD_DATE "] (" BUILD_BUILDUSER ")</span>"); 

    gtk_widget_realize (self->window);

    if (bmp_configuration_key_fetch_boolean (bmp_configuration, "bmp", "display-tooltips"))
	bmp_tooltips_enable (tooltips);
    else
	bmp_tooltips_disable (tooltips);

}

void
bmp_preferences_window_show(BmpWindowPreferences *self)
{
    static gboolean updated = FALSE;

    gtk_window_set_position (GTK_WINDOW (self->window), GTK_WIN_POS_CENTER_ALWAYS);
    gtk_window_present (GTK_WINDOW(self->window));

    if (!updated)
    {
        updated = TRUE;
        skin_view_update(self);
	gtk_widget_show_all(self->private->skin_view);

	/* FIXME: The FAM monitoring system has a HUGE memory leak somewhere */
	self->private->skins_path = g_build_filename(DATA_DIR,G_DIR_SEPARATOR_S,BMP_SKIN_DIR_BASENAME,NULL);

	fam_monitor_watch_new (bmp_fam_monitor,
				"skins",
				self->private->skins_path,
				(FamAddedFunc)skin_added_func,
				(FamRemovedFunc)skin_removed_func,
				self,
				NULL);

	fam_monitor_watch_new (bmp_fam_monitor,
				"user skins",
				BMP_PATH_USER_SKIN_DIR,
				(FamAddedFunc)skin_added_func,
				(FamRemovedFunc)skin_removed_func,
				self,
				NULL);

    }
}

gboolean
bmp_preferences_window_raise (BmpWindowPreferences *self)
{
	if (GTK_WIDGET_VISIBLE(self->window))
	{
		gdk_window_raise (self->window->window);
		return TRUE;
	}

	return FALSE;
}
